#include "hash.h"
#include "bitboards.h"
#include "parse.h"
#include "alfabeta.h"
#include <stdlib.h>
#include <stdio.h>

int64 CHANGE_COLOR;

struct transp_entry *table;
struct pawn_entry *pawn_table;
int64 zobrist[6][2][64];
int64 zobrist_castling[2][5];
int64 zobrist_passant[64];
unsigned char sequence;
extern bitboard square[64];

//Statistics for entire search
int rhits = 0;    //hits that return a score
int mhits = 0;    //hits that return a move
int misses = 0;
int probes = 0;

//Statistics for pawn search
int pawn_hits = 0;
int pawn_misses = 0;
int pawn_probes = 0;

//Statistics for quiescence search
int q_rhits = 0;    //hits that return a score
int q_mhits = 0;    //hits that return a move
int q_misses = 0;
int q_probes = 0;

/* In the postanalysis phase, if there is several best moves (with the same
   value), we pick the move that gives the best outcome after just 2 plies.
   The idea is to pick a move with a quick outcome rather than a gain farther
   away. In order to check that, we need to turn off probing of the
   transposition table, which is done by calling turn_off_hash(), which in
   turn sets the variable transp_probing to zero. */
int transp_probing = 1;

/* In the nullmove search we have to be careful with what we record in the
   transposition table. All values outside of the minimal window are invalid.
   We can store values that are outside of the interval we are searching,
   because those values don't need to be correct, and then they can be reused
   in subsequent searches. Values inside the interval we are searching
   (between alpha and beta) should not be stored however, because they will
   have the wrong value, and therefore cannot be reused in subsequent searches.
   However we can always safely store the best move. If the variable
   nullmove is set, the function record_hash uses the variables null_low and
   null_high to determine and interval where we can safely store position
   values. */
int nullmove = 0;
int null_low;
int null_high;

int64 rand64() {
  return rand() ^ ((int64) rand() << 15) ^ ((int64) rand() << 30)
    ^ ((int64) rand() << 45) ^ ((int64) rand() << 60);
}

void turn_on_hash() {
  transp_probing = 1;
}

void turn_off_hash() {
  transp_probing = 0;
}

void init_transptable() {
  int i, j, k;

  table = (struct transp_entry *) malloc(NBR_ENTRIES*sizeof(struct transp_entry));
  pawn_table = (struct pawn_entry *) malloc(PAWN_HASH_SIZE*sizeof(struct pawn_entry));
  sequence = 255;
  clear_hashtable();
  turn_on_hash();
  for (i = 0; i < 6; i++)
    for (j = 0; j < 2; j++)
      for (k = 0; k < 64; k++)
	zobrist[i][j][k] = rand64();
  zobrist_castling[WHITE][0] = rand64();
  zobrist_castling[WHITE][LONG_CASTLING_OK] = rand64();
  zobrist_castling[WHITE][SHORT_CASTLING_OK] = rand64();
  zobrist_castling[WHITE][LONG_CASTLING_OK | SHORT_CASTLING_OK] = rand64();
  zobrist_castling[WHITE][CASTLED] = rand64();
  zobrist_castling[BLACK][0] = rand64();
  zobrist_castling[BLACK][LONG_CASTLING_OK] = rand64();
  zobrist_castling[BLACK][SHORT_CASTLING_OK] = rand64();
  zobrist_castling[BLACK][LONG_CASTLING_OK | SHORT_CASTLING_OK] = rand64();
  zobrist_castling[BLACK][CASTLED] = rand64();
  for (i = 0; i < 64; i++)
    zobrist_passant[i] = rand64();
  CHANGE_COLOR = rand64();
  j = NBR_ENTRIES*sizeof(struct transp_entry);
  printf("size of transposition table = %d kb\n",j/1024);
  j = PAWN_HASH_SIZE*sizeof(struct pawn_entry);
  printf("size of pawn hashtable = %d kb\n",j/1024);
}

int64 get_zobrist_key_for_board(struct board *board, int color_to_move) {
  int64 zobrist_key = 0;
  bitboard pieces;
  int i, j, square_nbr;

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 6; j++) {
      pieces = board->piece[i][j];
      while (pieces != 0) {
	square_nbr = get_first_bitpos(pieces);
	zobrist_key ^= zobrist[j][i][square_nbr];
	pieces &= ~square[square_nbr];
      }
    }
  }
  zobrist_key ^= zobrist_castling[WHITE][board->castling_status[WHITE]];
  zobrist_key ^= zobrist_castling[BLACK][board->castling_status[BLACK]];
  zobrist_key ^= zobrist_passant[board->passant];
  return zobrist_key;
}

int64 get_zobrist_pawn_key_for_board(struct board *board) {
  int64 zobrist_key = 0;
  bitboard pieces;
  int i, square_nbr;

  for (i = 0; i < 2; i++) {
    pieces = board->piece[i][PAWN];
    while (pieces != 0) {
      square_nbr = get_first_bitpos(pieces);
      zobrist_key ^= zobrist[PAWN][i][square_nbr];
      pieces &= ~square[square_nbr];
    }
  }
  return zobrist_key;
}

void clear_hashtable() {
  int i;

  if (sequence < 255)
    sequence++;
  else {
    sequence = 1;
    for (i = 0; i < NBR_ENTRIES; i++) {
      table[i].lock = 0;
      table[i].sequence = 0;
    }
  }
}

/* After an aspiration search that returned a value outside of the
   alpha-beta bound, the values in the transposition table cannot be used.
   However the best moves can always be used. This function takes care
   of preparing the transposition table for an aspiration re-search. */
void prepare_research(int safe_low, int safe_high) {
  int i;

  for (i = 0; i < NBR_ENTRIES; i++) {
    if (table[i].value < safe_low || table[i].value > safe_high) {
      if (table[i].status & BEST_MOVE)
	table[i].status |= MOVE_ONLY;
      else
	table[i].lock = 0;
    }
  }
}

/* The argument can't be a normal int64, because that could give us
   a negative hash key. So we have to use unsigned. */
int get_hashkey(unsigned long long int zobrist_key, long long int hash_size) {
  return (zobrist_key & (hash_size-1));
}

int probe_pawn_hash(struct board *board, struct pawn_entry *pawn_struct) {
  struct pawn_entry *entry;

  pawn_probes++;
  entry = &(pawn_table[get_hashkey((unsigned long long int) board->zobrist_pawn_key,(long long int) PAWN_HASH_SIZE)]);
  if (entry->lock == board->zobrist_pawn_key) {
    pawn_hits++;
    *pawn_struct = *entry;
    return 0;
  } else {
    pawn_misses++;
    return UNKNOWN;
  }
}

int probe_hash(int depth, struct board *board, struct move *move,
	       int *alpha, int *beta, int *retval, int quies) {
  struct transp_entry *entry;
  //extern int re_search;
  //extern int prev_alpha;
  //extern int prev_beta;

  if (transp_probing) {
    probes++;
    if (quies)
      q_probes++;
    entry = &(table[get_hashkey((unsigned long long int) board->zobrist_key,(long long int) NBR_ENTRIES)]);
    if (entry->lock == board->zobrist_key) {
      /* Only take the hash value if it's at the same depth, or deeper than
	 the current depth, or if it is in quiescence search return the hash
	 value no matter what the depth is. */
      if ((quies && (entry->status & QUIESCENCE)) || ((!quies) && (entry->depth >= ((char)depth)) && ((entry->status & QUIESCENCE) == 0))) {
	rhits++;
	if (quies)
	  q_rhits++;
	
	if (!(entry->status & MOVE_ONLY)) {
	  /* If the aspiration search failed, then we can't trust values that are
	     outside of the original search window. If however we are outside the
	     alpha-beta bounds but on the "right" side (the side that didn't get
	     expanded) a cutoff should be safe. (?) */
	  /*if (re_search) {
	    if ((entry->status & EXACT)
		&& ((prev_alpha <= entry->value && entry->value <= prev_beta)
		    || (entry->value <= (*alpha) && (*alpha) == prev_alpha)
		    || (entry->value >= (*beta) && (*beta) == prev_beta)))
	      return entry->value;
	    / *if (entry->status & EXACT)
	      return entry->value;/
	    / *if (entry->status & ALPHA
	      && prev_alpha < entry->value && entry->value < prev_beta) {
	      *retval = entry->value;
	      *beta = min(*beta,*retval);
	      }/
	    if (entry->status & ALPHA) {
	      if (entry->value <= (*alpha) && (*alpha) == prev_alpha)
		return entry->value;
	      else if (prev_alpha <= entry->value && entry->value <= prev_beta) {
		*retval = entry->value;
		*beta = min(*beta,*retval);
	      }
	    }
	    / *if (entry->status & BETA
	      && prev_alpha < entry->value && entry->value < prev_beta) {
	      *retval = entry->value;
	      *alpha = max(*alpha,*retval);
	      }/
	    if (entry->status & BETA) {
	      if (entry->value >= (*beta) && (*beta) == prev_beta)
		return entry->value;
	      else if (prev_alpha <= entry->value && entry->value <= prev_beta) {
		*retval = entry->value;
		*alpha = max(*alpha,*retval);
	      }
	    }
	  } else {*/
	    if (entry->status & EXACT)
	      return entry->value;
	    if (entry->status & ALPHA) {
	      if (entry->value <= (*alpha)) {
		return entry->value;
	      } else {
		*retval = entry->value;
		*beta = min(*beta,*retval);
	      }
	    }
	    if (entry->status & BETA) {
	      if (entry->value >= (*beta))
		return entry->value;
	      else {
		*retval = entry->value;
		*alpha = max(*alpha,*retval);
	      }
	    }
	    //}
	}
	rhits--;
	if (quies)
	  q_rhits--;
      }
      if (entry->status & BEST_MOVE) {
	mhits++;
	if (quies)
	  q_mhits++;
	(*move).fsquare = square[(int) entry->bestmove_fsquare];
	(*move).tsquare = square[(int) entry->bestmove_tsquare];
	return HASH_MOVE;
      }
    }
    misses++;
    if (quies)
      q_misses++;
  }
  return UNKNOWN;
}

void record_pawn_hash(struct board *board, struct pawn_entry *pawn_struct) {
  struct pawn_entry *entry;

  entry = &(pawn_table[get_hashkey((unsigned long long int) board->zobrist_pawn_key,(long long int) PAWN_HASH_SIZE)]);
  pawn_struct->lock = board->zobrist_pawn_key;
  *entry = *pawn_struct;
}

void record_hash(int depth, struct board *board, struct move move,
		 int is_best_move, int value, int flag, int quies) {
  struct transp_entry *entry;

  entry = &(table[get_hashkey((unsigned long long int) board->zobrist_key,(long long int) NBR_ENTRIES)]);
  /* Only replace if same depth or deeper, or if the element pertains to
     an ancient search, or if it is quies we should always replace. */
  if (quies || ((!quies) && ((((char)depth) >= entry->depth) || (sequence > entry->sequence)))) {
    entry->lock = board->zobrist_key;
    if (is_best_move) {
      entry->status = BEST_MOVE;
      entry->bestmove_fsquare = get_first_bitpos(move.fsquare);
      entry->bestmove_tsquare = get_first_bitpos(move.tsquare);
    } else
      entry->status = 0;
    if (nullmove && (value < null_low || null_high < value))
      entry->status |= MOVE_ONLY;
    if (quies)
      entry->status |= QUIESCENCE;
    entry->value = value;
    entry->depth = (char) depth;
    entry->status |= (char) flag;
    entry->sequence = sequence;
  }
}


